private import lua;
private import lualib;
private import lauxlib;
private import std.string;

class Lua
{
public:
	// 使用するライブラリ
	enum
	{
		LIB_NONE   = 0x00, // 標準のみ
		LIB_MATH   = 0x01, // 数学関連
		LIB_STRING = 0x02, // 文字列操作
		LIB_TABLE  = 0x04, // テーブル操作
		LIB_IO     = 0x08, // I/O
	}
private:
	lua_State _lua; // Luaオブジェクト
	int       _lib; // 使うライブラリ
public:
	this()
	{
		_lua = null;
		_lib = LIB_NONE;
	}
	~this() { close(); }
	// 数値型を取得
	static int readInt(lua_State L, char[] name, int idx)
	{
		if(lua_type(L, idx) != LUA_TNUMBER)
		{
			throw new Error(format("%s(): Invalid parameter type. expected %d to NUMBER type", name, idx));
		}
		return cast(int)lua_tonumber(L, idx);
	}
	// 真偽型を取得
	static bool readBoolean(lua_State L, char[] name, int idx)
	{
		if(lua_type(L, idx) != LUA_TBOOLEAN)
		{
			throw new Error(format("%s(): Invalid parameter type. expected %d to BOOL type", name, idx));
		}
		return lua_toboolean(L, idx) != 0;
	}
	// 文字列型を取得
	static char[] readString(lua_State L, char[] name, int idx)
	{
		if(lua_type(L, idx) != LUA_TSTRING)
		{
			throw new Error(format("%s(): Invalid parameter type. expected %d to STRING type", name, idx));
		}
		return std.string.toString(lua_tostring(L, idx));
	}
	// 使うライブラリの指定
	void setUseLib(int param) { _lib = param; }
	// Luaオープン／ロード
	void load(char[] filepath)
	{
		close();
		// Lua生成
		_lua = lua_open();
		// ライブラリオープン
		lua_baselibopen(_lua);
		if(_lib & LIB_MATH)   luaopen_math(_lua);   // 数学関連
		if(_lib & LIB_STRING) luaopen_string(_lua); // 文字列操作
		if(_lib & LIB_TABLE)  luaopen_table(_lua);  // テーブル操作
		if(_lib & LIB_IO)     luaopen_io(_lua);     // I/O
		// Luaロード
		int ret = (luaL_loadfile(_lua, toStringz(filepath)));
		switch(ret)
		{
		case LUA_ERRFILE:
			throw new Error("Not exist : " ~ filepath);
		case LUA_ERRSYNTAX:
			throw new Error("Lua Syntax error : " ~ filepath);
		default:
			break;
		}
	}
	// 関数の登録
	void regist(char[] name, lua_CFunction f)
	{
		if(_lua is null) return;
		
		lua_register(_lua, toStringz(name), f);
	}
	// Luaクローズ
	void close()
	{
		if(_lua)
		{
			lua_close(_lua);
			_lua = null;
		}
	}
	// Lua再開
	void resume()
	{
		if(_lua)
		{
			// lua再開
			lua_resume(_lua, 0);
		}
	}
}
