private import hell2;
private import sysinfo;
private import token;

/**
 * パーティクル
 */
class Particle : Token
{
public:
	static TokenManager!(Particle) parent; // 親
	char[] id;
	int    timer;
	char[] msg;
	int[]  color;
	float  param1;
	float  param2;
	float  speed;  // スピード
public:
	void create(char[] id, float x, float y, float w, float h, float deg=0, float speed=0, int timer=0, int color[]=null, char[] msg=null, float ax=0f, float ay=0f)
	{
		this.id    = id;
		this.timer = timer;
		super.create(x, y, w, h, w, h, Hell_cosEx(deg)*speed, -Hell_sinEx(deg)*speed, ax, ay);
		this.color = color;
		this.msg   = msg;
	}

	static Particle add(char[] id, float x, float y, float w, float h, float deg=0, float speed=0, int timer=0, int[] color=null, char[] msg=null, float ax=0f, float ay=0f)
	{
		Particle particle = parent.getToken();
		if(!particle) return null;
		particle.create(id, x, y, w, h, deg, speed, timer, color, msg, ax, ay);
		return particle;
	}
	// プレイヤーブラー
	static void addBlur(float x, float y, int timer, float rot)
	{
		Particle particle = add("blur", x, y, 0, 0, 0, 0, timer);
		if(particle) { particle.rot = rot; }
	}
	// 破壊エフェクト(球)
	static void addDestroy(float x, float y, int[] color=[0xff, 0xff, 0xff])
	{
		add("destroy", x, y, 64, 64, 0, 0, 60, color);
		float rot = Hell_randint(0, 360);
		for(int i = 0; i < 5; i++)
		{
			int   w     = Hell_randint(24, 48);
//			float speed = Hell_randint(8, 64) / 10f;
			float speed = Hell_randint(16, 128)/10f;
			Particle particle = add("destroy", x, y, w, w, rot, speed, 60, color);
			rot += Hell_randint(48, 72);
		}
	}
	// 破壊エフェクト(円)
	static void addDestroy2(float x, float y, int[] color=[0xff, 0xff, 0xff])
	{
		Particle particle = add("destroy2", x, y, 92, 92, 0, 0, 60, color);
	}
	// 破壊エフェクト(球)※中心のみ
	static void addDestroy3(float x, float y, int[] color=[0xff, 0xff, 0xff])
	{
		add("destroy", x, y, 64, 64, 0, 0, 60, color);
	}
	static void addCircle(float x, float y, int[] color=[0xff, 0xff, 0xff])
	{
		add("circle", x, y, 64, 64, 0, 0, 60, color);
	}
	// スコア用文字エフェクト
	static void addNumber(float x, float y, float deg, float speed, int val1, int val2)
	{
		Particle particle = add("number", x, y, 0, 0, deg, speed, 90);
		if(particle)
		{
			particle.param1 = val1;
			particle.param2 = val2;
			particle.ax     = Hell_randint(-5, 5)/10f;
			particle.ay     = Hell_randint(-5, 5)/10f;
		}
	}
	// 上に進む文字列
	static void addMessage(float x, float y, char[] msg, int[] color)
	{
		add("message", x-msg.length*16/2, y, 0, 0, 90, 1, 60, color, msg);
	}
	// ぴょーんと飛び出す文字列
	static void addMessage2(float x, float y, char[] msg, int[] color, int val)
	{
		Particle p = add("message2", x-msg.length*16/2, y, 0, 0, Hell_randint(75, 105), 4f, 90, color, msg, 0, .1f);
		if(p)
		{
			p.param1 = val;
		}
	}
	// 打ち返し弾の当たりエフェクト
	static void addReflection(float x, float y)
	{
		for(int i = 0; i < 1; i++)
		{
			float deg   = Hell_randint(0, 359);
			float speed = Hell_randint(1, 8);
			int   timer = Hell_randint(32, 128);
			add("reflection", x, y, 0, 0, deg, speed, timer);
		}
	}
	// バリア破壊エフェクト
	static void addBarrier(float x, float y)
	{
		for(int i = 0; i < 8; i++)
		{
			float deg   = Hell_randint(0, 359);
			float speed = Hell_randint(3, 12);
			int   timer = Hell_randint(32, 128);
			add("barrier", x, y, 0, 0, deg, speed, timer);
		}
	}
	// ショットが撃てるエフェクト
	static void addShot(float x, float y)
	{
		add("shot", x, y, 0, 0, 0, 0, 60);
	}
	// 体当たりエフェクト
	static void addBlow(float x, float y, float rot)
	{
		Particle particle = add("blow", x, y, 0, 0, 0, 0, 60);
		if(particle) { particle.rot = rot; }
	}
	// 切り裂きエフェクト
	static void addCut(float x, float y, float deg, float speed, int[] color=[0xff, 0xff, 0xff])
	{
		Particle particle = add("cut", x, y, 0, 0, deg, speed, 60, color);
		if(particle) { particle.rot = deg; }
	}
	// レンズフレアの追加
	static void addFlare(float x, float y, float deg, float speed)
	{
		Particle particle = add("flare", x, y, 0, 0, deg, speed, 60);
	}
	void update()
	{
		switch(id)
		{
		case "number":
			ax = ax*0.9f;
			ay = ay*0.9f;
			vx = vx*0.97f;
			vy = vy*0.97f;
			timer--;
			if(timer < 1) exist = false;
			break;
		case "destroy":
			w -= 1;
			if(w < 1) w = 1;
			vx = vx*0.97f;
			vy = vy*0.97f;
			timer--;
			if(timer < 1) exist = false;
			break;
		case "destroy2":
			w += 5;
			timer--;
			if(timer < 1) exist = false;
			break;
		case "destroy3":
			w -= 1;
			if(w < 1) w = 1;
			rot -= 8;
			vx = speed*Hell_cosEx(rot);
			vy = speed*-Hell_sinEx(rot);
			timer--;
			if(timer < 1) exist = false;
			break;
		case "reflection":
		case "barrier":
		case "cut":
		case "flare":
			vx = vx*0.97f;
			vy = vy*0.97f;
			timer--;
			if(timer < 1) exist = false;
			break;
		case "message2":
			vx = vx*0.97f;
			if(timer < 20) vy = 0;
			timer--;
			if(timer < 1) exist = false;
			break;
		default:
			timer--;
			if(timer < 1) exist = false;
			break;
		}
		super.move();
		if(isOutOfWindow2()) exist = false;
	}
	void draw()
	{
		switch(id)
		{
		case "circle":
			Hell_setAlpha(HELL_ALPHA_ADD);
			float radius = (60-timer)*4;
			int   alpha  = 255*timer/60;
			Hell_drawCircle(x, y, radius, 3, color[0], color[1], color[2], alpha);
			break;
		case "blur":
			Hell_setAlpha(HELL_ALPHA_ADD);
			float d = Hell_sinEx(90*timer/64);
			Hell_drawTextureEx("player", x, y, 0, 0, 128, 128, 0.5f+2.0f*(1-d), 0.5f*d, rot, 0x00, 0xff, 0x00, timer*4);
			break;
		case "destroy":
		case "destroy3":
			if((timer < 30) && ((timer%4)<2) == 0) break;
			Hell_setAlpha(HELL_ALPHA_ADD);
			int alpha   = timer > 30 ? 255 : 255*timer/30;
			float scale = w/128f;
			Hell_drawTextureEx("effect", x, y, 0, 0, 128, 128, scale, scale, 0f, color[0], color[1], color[2], alpha);
			break;
		case "destroy2":
			Hell_setAlpha(HELL_ALPHA_ADD);
			float scale = w/128f;
			int alpha   = timer > 30 ? 0xff : 0xff*timer/30;
			Hell_drawTextureEx("effect", x, y, 128, 0, 128, 128, scale, scale, 0f, color[0], color[1], color[2], alpha);
			break;
		case "number":
			const int BASE = 45;
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			int val;
			/*
			if(timer < BASE) val = cast(int)param1;
			else
			{
				int t = BASE - (timer - BASE);
				val   = cast(int)(param1*t/BASE);
			}
			int digit = 0;
			int tmp   = val;
			while(tmp > 1)
			{
				digit++;
				tmp = tmp/10;
			}
			float scale = param1*param2/1000f;
			if(scale < .5f) scale = .5f;
			if(scale > 1.6f)  scale = 1.6f;
			int px    = cast(int)(x - scale*32f*digit);
			int py    = cast(int)(y - scale*48f);
			int alpha = cast(int)(255*(BASE-Hell_abs(timer-BASE))/BASE);
			if(param2 > 1)
			{
				Hell_drawFontEx(px, py, scale, 0xff, 0xff, 0xff, alpha, "%dx%d", val, cast(int)param2);
			}
			else
			{
				Hell_drawFontEx(px, py, scale, 0xff, 0xff, 0xff, alpha, "%d", val);
			}
			*/
			if(timer < BASE) val = cast(int)param2;
			else
			{
				int t = BASE - (timer - BASE);
				val   = cast(int)(param2*t/BASE);
				if(val < 1) val = 1;
			}
			float scale = 0.5f + .5f*param2/32;
			int alpha = cast(int)(255*(BASE-Hell_abs(timer-BASE))/BASE);
			int px    = cast(int)(x - scale*32f);
			int py    = cast(int)(y - scale*48f);
//			if(param2 == 1)
//			{
//				Hell_drawFontEx(px, py, scale, 0xff, 0xff, 0xff, alpha, "%d", cast(int)param1);
//			}
//			else
			{
				if(param2 >= 10) Hell_drawFontEx(px, py, scale, 0xff, 0xff, 0xff, alpha, "x%2d", val);
				else             Hell_drawFontEx(px, py, scale, 0xff, 0xff, 0xff, alpha, "x%d", val);
			}
			break;
		case "message":
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			int alpha = timer > 30 ? 255 : 255*timer/30;
			Hell_drawFont(msg, cast(int)x, cast(int)y, .7f, color[0], color[1], color[2], alpha);
			break;
		case "message2":
			int alpha = timer > 30 ? 255 : 255*timer/30;
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			float val   = param1/GH.ENEMY_BASE_DAMAGE*10;
			if(val < 1) return;
			float scale = 0.4f + 1.6f*param1/(GH.MAIN_RATIO_MAX*GH.ENEMY_BASE_DAMAGE);
			if(scale > 2) scale = 2f;
			Hell_drawFontEx(cast(int)x, cast(int)y, scale, color[0], color[1], color[2], alpha, "%d", cast(int)val);
			break;
		case "reflection":
			Hell_setAlpha(HELL_ALPHA_ADD);
			int alpha = timer > 16 ? 255 : (timer%8 < 4 ? 255 : 0);
			Hell_drawTriangle(x, y, 6, timer*13, 0, 0x00, 0xff, 0x00, alpha);
			break;
		case "barrier":
			Hell_setAlpha(HELL_ALPHA_ADD);
			int alpha = timer > 16 ? 255 : (timer%8 < 4 ? 255 : 0);
			Hell_drawTriangle(x, y, 6+timer, timer*4, 0, 0xff, 0x00, 0xff, alpha);
			break;
		case "shot":
			for(int i = 0; i < 3; i++)
			{
				int t = timer;
				t -= i * 10;
				if(t < 0) break;
				int radius = 32 + 32 * (60-t)/60;
				int alpha  = 0xff * t/60;
				Hell_drawCircle(x, y, radius, 1, 0x40, 0xff, 0x40, alpha);
			}
			break;
		case "blow":
			Hell_setAlpha(HELL_ALPHA_ADD);
			int alpha   = timer > 30 ? 0x80 : 0x80*timer/30;
			Hell_drawTextureEx("effect", x, y, 0, 128, 256, 128, 1f, 1f, rot+45,  0xff, 0x00, 0x00, alpha);	
			Hell_drawTextureEx("effect", x, y, 0, 128, 256, 128, 1f, 1f, rot+135, 0xff, 0x00, 0x00, alpha);	
			break;
		case "cut":
			Hell_setAlpha(HELL_ALPHA_ADD);
			int alpha = timer > 30 ? 0xff : 0xff*timer/30;
			Hell_drawTextureEx("effect", x, y, 0, 128, 256, 128, 3f, 3f, rot+timer*4, color[0], color[1], color[2], alpha);	
			break;
		case "flare":
			Hell_setAlpha(HELL_ALPHA_ADD);
			int alpha = timer > 30 ? 0x80 : 0x80*timer/30;
			Hell_drawTextureEx("shot", x, y, 0, 0, 256, 256, 2f, 2f, rot+timer*4, 0xff, 0xff, 0xff, alpha);
			break;
		}
	}
}
